pub fn blockDataTransfer(comptime InstrFn: type, comptime P: bool, comptime U: bool, comptime S: bool, comptime W: bool, comptime L: bool) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u32) void {
            const rn: u4 = @truncate(opcode >> 16 & 0xF);
            const rlist: u16 = @intCast(opcode & 0xFFFF);

            const reg_count: u32 = @popCount(rlist);
            const first_in_list: u4 = @truncate(@ctz(rlist)); // note that @ctz(0x0000) and @ctz(0x0001) collide

            // U determines whether the LDM/STM transfer is made upwards (U == 1)
            // or downwards (U == 0).

            const start_addr: u32 = if (U) blk: {
                break :blk cpu.r[rn] + if (P) 4 else 0;
            } else blk: {
                break :blk cpu.r[rn] - (4 * reg_count) + if (!P) 4 else 0;
            };

            // FIXME : why 4 * reg_count?

            const new_base_addr: u32 = if (U) blk: {
                break :blk cpu.r[rn] + 4 * reg_count;
            } else blk: {
                break :blk cpu.r[rn] - 4 * reg_count;
            };

            var address = start_addr;

            // On Empty List:
            //
            // ARMv4 Only: R15 is Loaded/Stored
            // ARMv4/Armv5: Rn = Rn +/- 0x40

            if (rlist == 0) {
                if (Arm32.arch == .v4t) {
                    const undefined_addr: u32 = if (U) blk: {
                        break :blk cpu.r[rn] + if (P) 4 else 0;
                    } else blk: {
                        break :blk cpu.r[rn] - (0x40 - if (!P) 4 else 0);
                    };

                    if (L) {
                        cpu.r[15] = cpu.read(u32, undefined_addr);
                        cpu.pipe.reload(cpu);
                    } else {
                        cpu.write(u32, undefined_addr, cpu.r[15] + 4);
                    }
                }

                cpu.r[rn] = if (U) cpu.r[rn] + 0x40 else cpu.r[rn] - 0x40;
                return;
            }

            // What happens when W is set and Rn is in the rlist? (STM)
            //
            // Armv4: Store OLD Base if Rb is FIRST entry in Rlist, otherwise store NEW base
            // Armv5: Always store OLD Base

            // FIXME: This absolutely needs revisiting :skull:

            const r15_present = rlist >> 15 & 1 == 1;
            var write_to_base = true;

            for (first_in_list..16) |idx| {
                const i: u4 = @intCast(idx);

                if (rlist >> i & 1 == 1) {
                    transfer(cpu, r15_present, i, address);
                    address += 4;

                    if (W and !L and write_to_base) {
                        cpu.r[rn] = new_base_addr;
                        write_to_base = false;
                    }
                }
            }

            // What happens when W is set and Rn is in the rlist? (LDM)
            //
            // ARMv4: No writeback
            // ARMv5: writeback if Rn is "the ONLY register" or NOT the LAST register

            if (W and L) {
                if (rlist >> rn & 1 == 0) {
                    cpu.r[rn] = new_base_addr;
                    return;
                }

                switch (Arm32.arch) {
                    .v4t => {}, // No Writeback
                    .v5te => {
                        const rn_is_last = (15 - @clz(rlist)) <= rn;

                        if (reg_count == 1 or !rn_is_last) {
                            cpu.r[rn] = new_base_addr;
                        }
                    },
                }
            }
        }

        fn transfer(cpu: *Arm32, r15_present: bool, i: u5, address: u32) void {
            if (L) {
                if (S and !r15_present) {
                    // Always Transfer User mode Registers
                    cpu.setUserModeRegister(i, cpu.read(u32, address));
                } else {
                    const value = cpu.read(u32, address);

                    cpu.r[i] = value;
                    if (i == 0xF) {
                        cpu.r[i] &= ~@as(u32, 3); // Align r15
                        cpu.pipe.reload(cpu);

                        if (S) cpu.setCpsr(cpu.spsr.raw);
                    }
                }
            } else {
                if (S) {
                    // Always Transfer User mode Registers
                    // This happens regardless if r15 is in the list
                    const value = cpu.getUserModeRegister(i);
                    cpu.write(u32, address, value + if (i == 0xF) 4 else @as(u32, 0)); // PC is already 8 ahead to make 12
                } else {
                    cpu.write(u32, address, cpu.r[i] + if (i == 0xF) 4 else @as(u32, 0));
                }
            }
        }
    }.inner;
}
