const std = @import("std");
const rotr = @import("zba-util").rotr;
const PSR = @import("../../../arm.zig").PSR;

const log = std.log.scoped(.ctrl_ext_space);

pub fn control(comptime InstrFn: type, comptime I: bool, comptime op: u6) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u32) void {
            if (I) {
                // MSR (register)
                const R = op >> 5 & 1 == 1;
                msr(R, I, cpu, opcode);

                return;
            }

            switch (op) {
                0b00_0000, 0b10_0000 => { // MRS
                    const R = op >> 5 & 1 == 1;
                    const rd = opcode >> 12 & 0xF;

                    if (R and !cpu.hasSPSR()) log.err("Tried to read SPSR from User/System Mode", .{});
                    cpu.r[rd] = if (R) cpu.spsr.raw else cpu.cpsr.raw;
                },
                0b01_0000, 0b11_0000 => { // MSR (register)
                    const R = op >> 5 & 1 == 1;
                    msr(R, false, cpu, opcode);
                },
                else => cpu.panic("unhandled instruction: 0x{X:0>8}", .{opcode}),
            }
        }

        inline fn msr(comptime R: bool, comptime imm: bool, cpu: *Arm32, opcode: u32) void {
            const field_mask: u4 = @truncate(opcode >> 16 & 0xF);
            const rm_idx = opcode & 0xF;
            const right = if (imm) rotr(u32, opcode & 0xFF, (opcode >> 8 & 0xF) * 2) else cpu.r[rm_idx];

            if (R and !cpu.hasSPSR()) log.err("Tried to write to SPSR in User/System Mode", .{});

            if (R) {
                // arm.gba seems to expect the SPSR to do somethign in SYS mode,
                // so we just assume that despite writing to the SPSR in USR or SYS mode
                // being UNPREDICTABLE, it just magically has a working SPSR somehow
                cpu.spsr.raw = fieldMask(&cpu.spsr, field_mask, right);
            } else {
                if (cpu.isPrivileged()) cpu.setCpsr(fieldMask(&cpu.cpsr, field_mask, right));
            }
        }
    }.inner;
}

fn fieldMask(psr: *const PSR, field_mask: u4, right: u32) u32 {
    // This bitwise ORs bits 3 and 0 of the field mask into a u2
    // We do this because we only care about bits 7:0 and 31:28 of the CPSR
    const bits: u2 = @truncate((field_mask >> 2 & 0x2) | (field_mask & 1));

    const mask: u32 = switch (bits) {
        0b00 => 0x0000_0000,
        0b01 => 0x0000_00FF,
        0b10 => 0xF000_0000,
        0b11 => 0xF000_00FF,
    };

    return (psr.raw & ~mask) | (right & mask);
}
